/**
 * \file  dslhal_debug_functions.h
 *        header file for DSL HAL Debug Application.
 * \info  DSL Systems Software Support
 *        Copyright (c) 2008 Infineon Technologies.
 *        All Rights Reserved
 *
 * \date  02Jul08    0.00.00     CZ   SMS00784491. merged
  **************************************************************************/

#ifndef __DSLHAL_DEBUG_FUNCTIONS_H__
#define __DSLHAL_DEBUG_FUNCTIONS_H__

#define askHelp() (argv[1][0]=='*')

#define ADSL2_MASKS  (ADSL2_ANNEX_A_OR_B | ADSL2_ANNEX_I | ADSL2_ANNEX_J | ADSL2_ANNEX_L | ADSL2_ANNEX_M)
#define ADSL2PLUS_MASKS (ADSL2PLUS_ANNEX_A_OR_B | ADSL2PLUS_ANNEX_I | ADSL2PLUS_ANNEX_J | ADSL2PLUS_ANNEX_M)

#define NO_MODE                       0x0000
#define DELT_ENABLE                   0x0001
#define GDMT_ANNEX_A_OR_B             0x0002
#define GLITE_ANNEX_A_AND_B           0x0004
#define ADSL2_ANNEX_A_OR_B            0x0008
#define ADSL2PLUS_ANNEX_A_OR_B        0x0010
#define ADSL2_ANNEX_L                 0x0020
#define T1413_ANSI                    0x0080
#define ADSL2_ANNEX_I                 0x0100
#define ADSL2_ANNEX_J                 0x0200
#define ADSL2_ANNEX_M                 0x0400
#define ADSL2PLUS_ANNEX_I             0x0800
#define ADSL2PLUS_ANNEX_J             0x1000
#define ADSL2PLUS_ANNEX_M             0x2000
#define ADSL2_ANNEX_A_OR_B_DELT       (ADSL2_ANNEX_A_OR_B     + DELT_ENABLE)
#define ADSL2PLUS_ANNEX_A_OR_B_DELT   (ADSL2PLUS_ANNEX_A_OR_B + DELT_ENABLE)
#define ADSL2_ANNEX_L_DELT            (ADSL2_ANNEX_L          + DELT_ENABLE)
#define OLDMULTI_MODE                 (GDMT_ANNEX_A_OR_B+GLITE_ANNEX_A_AND_B+T1413_ANSI)
#define MULTI_MODE                    0xffff
#define MULTI_MODE8                   0xff        // 8 bit Multi_Mode for backward compatibility


#define DEV_HOST_DSP_OAM_INTERNAL_POINTER_LOCATION 0x80000004

typedef signed short * PSINT16;
typedef signed int * PSINT32;
typedef unsigned char * PUINT8;
typedef unsigned int *PUINT32;
typedef int *PINT32;
typedef signed char SINT8;
typedef signed short SINT16;
typedef UINT8      UChar;
typedef UINT8      UInt8;
typedef UINT16     UInt16;
typedef UINT32     UInt32;
typedef UINT32     DWORD;

/// Bit definition for MainControl.showTimeDispControl
#define SHOWTIME_DISP_BITALLOC_DS   0x000001
#define SHOWTIME_DISP_BITALLOC_US   0x000002
#define SHOWTIME_DISP_DSPERROR      0x000004
#define SHOWTIME_DISP_DSMARGIN      0x000008


#define DS_RX_CODEWORD 1 ///< fro US_FRAME
#define DS_LOOP_BACK  0  ///< for DDI_TX_MSG below
#define SHALF  1         ///< for DDI_RX_MSG below


//-------- Following from ctrl_msm_swif.h -------------------------------------
#if DS_LOOP_BACK
#define RX_LATBUFSIZE                        6024
#else
#define MAX_SP_FACTOR                        4
#define MAX_SP_FACTOR_STD                    3
#define MAX_CODEWORD_SIZE                    256
#define RX_LATBUFSIZE                        ((MAX_SP_FACTOR+1)*MAX_CODEWORD_SIZE)
#endif
//--------------- Followings from di_bits_swif.h -------------------------------
#define RS_N_max       255 // Maximum size for RS code word
#if SHALF
// Maximum size for DS K, make K+1 to be DWORD aligned  // (dkc)
#define K_DS_max      (MAX_SP_FACTOR * MAX_CODEWORD_SIZE - 1)
#else
#define K_DS_max       255 // Maximum size for DS K, make K+1 to be DWORD aligned  // (dkc)
#endif
#define K_US_max       95  // Maximum size for US K => make K+1 to be DWORD aligned

// The frame structure ready for bit-extraction
typedef  struct
{
  UChar   FastByte;
  UChar   ChData[K_DS_max];
} DS_FRAME;
typedef  struct
{
  UChar   FastByte;
#if DS_LOOP_BACK
  UChar   ChData[K_DS_max];
#else
#if DS_RX_CODEWORD
  UChar   ChData[RS_N_max]; // Was ChData[K_DS_max] Max size of US_FRAME can be atmost 1 codeword size
#else
  UChar   ChData[K_US_max];
#endif
#endif
} US_FRAME;

typedef struct{
  int     re;
  int     im;
} complex_int, *PTR_complex_int;

// Overhead info bytes
typedef struct
{
  UInt8 crc;
  UInt8 ib0_to_7;
  UInt8 ib8_to_15;
  UInt8 ib16_to_23;
  UInt8 ib24_to_31;
  UInt8 eoc1_to_6;
  UInt8 eoc7_to_13;
  UInt8 aoc;
  UInt8 dummy;
} OVHDVARS;


// Bit Swap message
typedef struct
{
  UInt8 cmnd;
  UInt8 toneindex;
} FIELD;

typedef struct
{
  UInt8 header;
  UInt8 dummy;
  UInt16 dummy2;
  FIELD Field[4];
} BITSWAPREQ;

typedef struct {
  UINT8         lastMp;                                                                   
  PUINT8        wrPtr;                    
  PUINT8        rdPtr;                    
  UINT8         callBitbyteChainFlag;     
  PUINT8        endofArrayPtr;            
  PUINT8        startofArrayPtr;          
  UINT8         usedBits;                 
  UInt8         resetEndofArrayPtr;       
  UINT8         resetEvenBytes;                                                 
  UINT16        bits;                     
} DI_LMUX_parameters_t;

typedef struct
{
  US_FRAME      frame;              
  OVHDVARS      OvhdVars;           
  UInt8         intlv;              
  UInt8         fr_per_sframe;      
#if DS_LOOP_BACK
  UInt16        tone_num;           
#else
  UInt8         tone_num;           
#endif
  UInt8         start_tone;         
  UInt8         n;                  
  UInt8         k;                  
  UInt8         r;                  
  UInt8         D;                  
  UInt8         s;                 
  UInt8         frame_cntr;         
  UInt8         bitswap;           
  UInt8         dummy;
  BITSWAPREQ    BitSwapReq;         
  UINT8         MSGc;               
  UINT8         Tp;                 
  UINT16        Lp;                 
  DI_LMUX_parameters_t   DI_LMUX_parameters_s; 
  UINT8         Tp_cntr;            
  UINT8         SEQp_cntr;          
  UINT16        numTCBytes;         
  OVHDVARS      adsl2OvhdVars;      
} DDI_TX_MSG;


typedef struct
{
  DS_FRAME      frame;              
  OVHDVARS      OvhdVars;           
  UInt8         intlv;              
  UInt8         fr_per_sframe;      
  UInt16        tone_num;           
  UInt16        start_tone;         
  UInt8         rx_index;           
  UInt8         r;                  
  UInt16        D;                  
  UInt16        k;                  
  UInt16        n;                  
  UInt8         s;                  
#if SHALF
  UInt8         rx_s_half;          
  UInt8         Nflag;              
  UInt8         gap;                
  UInt16        K1;                 
  UInt16        K2;                 
  UInt16        N1;                 
  UInt16        N2;                 
  UInt16        DMAcountForCodeWord0; 
  UInt16        DMAcountForCodeWord1; 
#else
  UInt16        N1;                   
  UInt16        DMAcountForCodeWord0; 
#endif

  UInt8         crc_err;            
  UInt8         FEC_err;            
  // DS CRC counters
  UInt32        ds_crc_errcnt;      
  UInt32        ds_crc_gdcnt;       
  // DS FEC counters
  UInt32        ds_fec_errcnt;      
  UInt32        ds_fec_gdcnt;       
  UInt32        RsNotDoneErrCnt;    

  // Counters
  UInt8         extra_delay;        
  UInt16        delay;              
  UInt8         rx_s_frame_cntr;    
  UInt8         Rx_Steady_State_Flag;    
  UInt32        rx_superframe_cntr; 

  UInt8         frame_cntr;         
  UInt8         bitswap;            
  UInt16        dummy;              
  BITSWAPREQ    BitSwapReq;         
  UINT8         MSGc;               
  UINT8         Tp;                 
  UINT16        Lp;                 
  DI_LMUX_parameters_t   DI_LMUX_parameters_s; 
  UINT8         Tp_cntr;            
  UINT8         SEQp_cntr;          
  UINT16        bytesJunked;        
  UInt16        numTCBytes;         
  UInt8         numDecCwords;       
  UInt8         spanLessThanOne;    
  UInt8         rsCwdDecDone;       
  UInt8         bytesNotJunked;     
  OVHDVARS      adsl2OvhdVars;      
} DDI_RX_MSG;

/// Structure used to present AGC settings to host.
typedef struct
{
  UINT8 rxPga1; ///< receive power gain 1
  UINT8 rxPga2; ///< receive power gain 2
  UINT8 rxPga3; ///< receive power gain 3
  UINT8 anlgEq; ///< analog equalizer
} DEV_HOST_agcSetting_t;

typedef struct
{
  UINT32 realTimeErrors;             ///< counts number of real time errors
} CTRL_MAIN_exceptionCounts_t;

typedef struct
{
  PUINT32 realTimeErrors_p;    ///< if (realTimeErrors > 0), DSP has blown real-time
} DEV_HOST_dspErrorCounts_t;

/// environment struct
typedef struct
{
  volatile UINT32 debugHalt;            ///< Only used if concurrent debugging between CCS and atmdiag -j(tag)
  PSINT16         snr_p;                ///< pointer to snr table
  PTR_complex_int channelId_p;          ///< pointer to channel id
  PSINT32         doubleXmtRate_p;      ///< Pointer to modem[0].double_xmt_rate
  UINT32          atucId;               ///< bit 0: TI/bit 1: Acaltel/bit 2: GPSN/bit 3: ADI
  UINT8           testMode;             ///< DSP in test mode, DDI data will be bad
  UINT8           txSwitchLogic;        ///< Set to "1" to enable 32/64 auto detect
  UINT8           targetMarginAdjust;   ///< DS target margin adjustment flag
  UINT8           reserved1;            ///< dummy byte for 32-bit alignment
  PSINT32         doubleRxRate_p;       ///< This flag indicates how many tones are used in the DS direction.
  UINT8           ghsVendorId[8];       ///< G.hs Vendor ID
  UINT32          currCoProfCtrlAddr;   ///< Address of Co Profile Control currently in use in DSP
  UINT32          currCoProfAddr;       ///< Address of Co Profile currently in use in DSP
  UINT32          ghsATUCVendorIdAddr;  ///< Address of IMSG_HS_rxIdField_s.vendorID
  UINT32          ovhdControl;          ///< enable/control overhead function
  UINT8           ProgCode[4];          ///< Progress code
  UINT32          TestParmHlogfMsgAddr; ///< Address of TestParmHlogfMsg[]
  UINT32          TestParmQLNfMsgAddr;  ///< Address of TestParmQLNfMsg[]
  UINT32          TestParmSNRfMsgAddr;  ///< Address of TestParmSNRfMsg[]
  UINT32          recvMsgAddr;          ///< Address of recv_msg, type of DDI_RX_MSG
  UINT32          xmtMsgAddr;           ///< Address of xmt_msg, type of DDI_TX_MSG

  UINT32          L1CapCtrlAddr;
} DEV_HOST_modemEnv_t;


typedef struct
{
  UINT32 overlayHostAddr;     ///< source address in host memory
  UINT32 overlayXferCount;    ///< number of 32bit words to be transfered
  UINT32 overlayDspAddr;      ///< destination address in DSP's PMEM
} DEV_HOST_olayDpPageDef_t;


#define DEV_HOST_PAGE_NUM 11  ///< number of overlay pages (page 0 + number of overlay pages)

typedef struct
{
  UINT32 overlayStatus;       ///< Status of current overlay to DSP PMEM
  UINT32 overlayNumber;       ///< DSP PMEM overlay page number
  UINT32 overlayState;        ///< Indicates whether current state is an overlay state
  DEV_HOST_olayDpPageDef_t *olayDpPage_p[DEV_HOST_PAGE_NUM]; ///< Def's for the Pages
} DEV_HOST_olayDpDef_t;

#define MAX_CMSGPCB2_LENGTH              12

///struct for negotiotiation parameters
typedef struct
{
  UINT8  trainMode;          ///< Train mode selected.  See training modes defined above.
  UINT8  bDummy1;            ///< dummy byte for explicit 32-bit alignment
  UINT16 lineLength;         ///< Contains loop length estimate (26AWG line length equivalent).
  UINT32 atucVendorId;       ///< Pass the vendor id of the CO to the host
  UINT8  cMsgs1[MAX_CMSGPCB2_LENGTH];    // Array is used for both cMsgs1 & cMsg-PCB
  UINT16 adsl2DSRate;        ///<
  UINT8  cRates2;            ///<
  UINT8  rRates2;            ///<
  UINT8  rRates1[4][11];     ///<
  UINT8  cMsgs2[4];          ///<
  UINT8  cRates1[4][30];     ///<
  UINT8  rMsgs2[4];          ///< This always needs to be aligned on 4 byte boundary.
  UINT16 adsl2USRate;        ///<
  UINT8  atucGhsRevisionNum; ///< Pass the G.hs Revision number of the CO to the host
  UINT8  reserved1;          ///<
  PUINT8 *adsl2DeltMsgs_p;   ///< This pointer to a pointer passes the address of the globalvar.pString, which is also
                             ///< a pointer list of pointers. It will be used to pass all the new ADSL2 DELT messages to
                             ///< host side. This is for ACT.
  UINT8  rcvMode;            ///< rcvMode for physical layer transactions. Trainmode will be used for all messaging
  UINT8  bDummy2;            ///< for integer alignment
  UINT16 annex_selected;     ///< Annex that the modem trained up with.
  UINT16 psd_mask_qualifier; ///< Bit definitions given above in the files.
  SINT16 raOVRate;           // CQ11194 HZ Store the overhead data rate calculated at training time. 
  UINT16 lineLength24AWG;    ///< Contains 24AWG equivalent of loop length estimate.
} DEV_HOST_dspWrNegoParaDef_t;
// The two values below can never change (for backwards compatibility of host intf.)
#define DEV_HOST_EOCREG4LENGTH   32
#define DEV_HOST_EOCREG5LENGTH   32

///struct for eoc variables
typedef struct
{
  UINT8        eocReg4[DEV_HOST_EOCREG4LENGTH];   ///< Host/Dsp Write, vendor specific EOC Register 4
  UINT8        eocReg5[DEV_HOST_EOCREG5LENGTH];   ///< Host/Dsp Write, vendor specific EOC Register 5
  UINT8        vendorId[8];               ///< Host write
  UINT8        revNumber[4];              ///< Host, ATU-R Revision Number
  UINT8        serialNumber[32];          ///< Host write
  UINT8        eocReg4Length;             ///< Host Write, valid length for EOC register 4
  UINT8        eocReg5Length;             ///< Host Write, valid length for EOC register 5
  UINT8        selfTestResults[2];        ///< EOC selftestResults place holder
  UINT32       eocModemStatusReg;         ///< Dsp Write, status bits to host
  // UR8_MERGE_START CQ11247_TR69_DS_LATN_SATN  YW
  UINT16       lineAtten;               // Dsp Write, line attenuation in 0.1 db step
  SINT16        dsMargin;                  ///< DSP Write, measured DS margin
  // UR8_MERGE_END CQ11247_TR69_DS_LATN_SATN  YW
  UINT8        aturConfig[30];            ///< Dsp Write, also used by EOC for ATUR Configuration
  // UR8_MERGE_START_END CQ11247_TR69_DS_LATN_SATN  YW
  SINT8        dummy[2];               // 32-bit alignment   // UR8_MERGE_START_END CQ11007 Ram
} DEV_HOST_eocVarDef_t;

///struct for clear eoc variables
typedef struct
{
  UINT16       endEocThresh;              ///< Host Write, end of Clear EOC stream threshold
  UINT16       dummy;                     ///< dummy value to fill gap
  UINT32       dropEocCount;              ///< Dsp Write, counter of dropped Clear EOC bytes
  UINT16       eocRxLength;               ///< Host/DSP write, number of valid Rx Clear EOC bytes
  UINT16       eocTxLength;               ///< Host/DSP write, number of valid Tx Clear EOC bytes
  UINT8        eocRxBuf[64];              ///< Dsp Write, Buffer for receiving Rx Clear EOC bytes
  UINT8        eocTxBuf[64];              ///< Host Write, Buffer for writing Tx Clear EOC bytes
} DEV_HOST_clearEocVarDef_t;


///struct for the dsp version
typedef struct
{
  UINT32 timestamp;    ///< Number of seconds since 01/01/1970
  UINT8  major;        ///< Major "00".00.00.00 revision nomenclature
  UINT8  minor;        ///< Minor 00."00".00.00 revision nomenclature
  UINT8  bugFix;       ///< Bug Fix 00.00."00".00 revision nomenclature
  UINT8  buildNum;     ///< Build Number 00.00.00."00" revision nomenclature
  UINT8  netService;   ///< Network service identifier
  UINT8  chipsetGen;   ///< chipset generation
  UINT8  chipsetId;    ///< chipset identifier
  UINT8  chipsetId2;   ///< second byte for "RV" chipset et al.
  UINT8  hardwareRev1; ///< hardware revision, 1st char
  UINT8  hardwareRev2; ///< hardware revision, 2nd char
  UINT8  featureCode;  ///< feature code
  UINT8  dummy1;       ///< dummy byte for explicit 32-bit alignment
} DEV_HOST_dspVersionDef_t;


///struct for the OAM negotiation parameters
typedef struct
{
  UINT8   stdMode;              ///< Desired train mode.  See training modes defined above.
  UINT8   ghsSequence;          ///< Selected G.hs session as shown in Appendix 1
  UINT8   usPilotFlag;          ///< Value of 1 indicates transmit an upstream pilot on bin 16
  UINT8   stdMode_byte2;        ///< 2nd byte of the desired train mode. This is to be used in conjunction with stdMode.
  UINT8   rMsgs1[38];           ///< RMSG-1(6) and RMSG_PCB (38)
  UINT16  stdMode_bytes34;      ///< Reserved for future use
  UINT32  oamFeature;           ///< 32 bit wide bit field to set OAM-specific features.
  SINT8   marginThreshold;      ///< Threshold for margin reporting
  UINT8   hostFixAgc;           ///< flag to force datapump to bypass AGC training and use the following values
  UINT8   hostFixEqualizer;     ///< forced analog equalizer value used during AGC training when hostfix_agc is on
  UINT8   hostFixPga1;          ///< forced pga1 value used during AGC training when hostFixAgc is on
  UINT8   hostFixPga2;          ///< forced pga2 value used during AGC training when hostFixAgc is on
  UINT8   hostFixPga3;          ///< forced pga3 value used during AGC training when hostFixAgc is on
  UINT8   marginMonitorShwtme;  ///< margin monitoring flag (during showtime)
  UINT8   marginMonitorTrning;  ///< margin monitoring flag (during training)
  UINT8   disableLosAlarm;      ///< flag to disable training based on los
  UINT8   usCrcRetrain;         ///< flag to disable retrain due to excessive USCRC
  UINT8   t1413VendorId[2];     ///< Vendor ID used for T1.413 trainings
  UINT8   gdmtVendorId[8];      ///< Vendor ID used for G.dmt trainings (ITU VendorID)
  UINT8   missingTones[64];     ///< 64 element array to define missing tones for TX_MEDLEY and TX REVERB tests
  UINT32  missingToneDsAddr;    ///< Address given to DSP for tones to be switched off in DS direction
  UINT8   dsToneTurnoff_f;      ///< This flag controls the DS tone turn off logic
  UINT8   mhzFlag;              ///< Indicates whether DSP runs at 200 v/s 250 Mhz
  UINT8   reserved2;            ///< Dummy bytes
  UINT8   reserved3;            ///< Dummy bytes
  UINT32  mode_config;          ///< This flag tells the DSP the modes which are enabled by OAM.
} DEV_HOST_oamWrNegoParaDef_t;


///struct for profile
typedef struct
{
//  DEV_HOST_coData_t             * hostProfileBase_p;                    // base address of profile list
  UINT32                        * hostProfileBase_p;       ///< \note base address of profile list
  UINT32                        * dspScratchMem_p0;        ///< \note base address of DSP scratch memory
  UINT32                        * dspScratchMem_p1;        ///< \note base address of DSP scratch memory
  UINT32                        * dspScratchMem_p2;        ///< \note base address of DSP scratch memory
  UINT32                        * dspScratchMem_p3;        ///< \note base address of DSP scratch memory
  //UR8_MERGE_START_END CQ11922 Tim
  UINT32                        * dspScratchMem_p4;        ///< \note base address of DSP scratch memory
} DEV_HOST_profileBase_t ;

enum
{
  REASON_LOS         = 0x01,
  REASON_DYING_GASP  = 0x02,
  REASON_USCRCERR    = 0x04,
  REASON_MARGIN_DROP = 0x08
};

///struct for modem environment
typedef struct
{
  SINT16 subStateIndex;        ///< Index that signifies datapump substate. (DSP write)
  UINT8  externalBert;         ///< Turn on/off external BERT interface. 0 = OFF; 1 = ON. (Host write)
  UINT8  usBertPattern;        ///< BERT pattern for US TX data. 0 = 2^15-1; 1 = 2^23-1. (Host write)
  UINT8  overrideDslLinkLed_f; ///< Overrides DSP operation of the DSL_LINK LED. (Host write)
                               ///   0 = DSP is in control; 1 = Host is in control.
  UINT8  dslLinkLedState_f;    ///< DSL_LINK LED state when override flag has been set. (Host write)
                               ///   DSL_LINK LED will be updated with this value once per frame.
                               ///   LED is active-low: 0 = ON, 1 = OFF.
  UINT8  overrideDslActLed_f;  ///< Overrides DSP operation of the DSL_ACT LED. (Host write)
                               ///   0 = DSP is in control; 1 = Host is in control.
  UINT8  dslActLedState_f;     ///< DSL_ACT LED state when override flag has been set. (Host write)
                               ///   DSL_ACT LED will be updated with this value once per frame.
                               ///   LED is active-low: 0 = ON, 1 = OFF.
  UINT8  dGaspLprIndicator_f;  ///< How LPR signal (GPIO[0]) is to be interpreted. (Host write)
                               ///   0 = LPR is active-low; 1 = LPR is active-high.
  UINT8  overrideDspLprGasp_f; ///< Overrides DSP detection of LPR signal to send out DGASP. (Host write)
                               ///   0 = DSP detects LPR; 1 = Host detects LPR and sends "HOST_DGASP" to DSP.
  UINT8  pcbEnabled;           ///< DS power cut back
  UINT8  maxAvgFineGainCtrl_f; ///< If maxAvgFineGainCtrl_f == 0, then the datapump controls the maximum average fine gain value.
                               ///    If maxAvgFineGainCtrl_f == 1, then the host controls the maximum average fine gain value.
  UINT32 reasonForDrop;        ///< This field will tell the host what might be the reason for a dropped connection.
  SINT16 maxAverageFineGain;   ///< When maxAvgFineGainCtrl_f == 1, the value in maxAverageFineGain is the maximum average fine gain level in 256log2 units.
  UINT8  infineon_CO_cabinet_mode_enabled; /// 0 Default
                                           /// 1 (Mode is G.992.5 mode) && (CO is Infineon) && (Cabinet mode with DS start tone # >= 100)
  UINT8  enable_cnxt_clrEoc_patch_f;
} DEV_HOST_modemEnvPublic_t;



//'@' debug cmd function prototypes
/**
 * This function called by \b \@snrd displays the signal to noise (SNR) during training
 * \param[in]  argc number of arguments
 * \param[in]  argv[0] 'snrd'
 * \param[in]  argv[1] 0=single path (default)
 * \param[out] argv[2] 0=dec (default)
 * \retval     0 error
 * \retval     1 success
 */
int dbgcmd_SNR_Ds (int argc, unsigned char **argv);


/**
 * This function called by \b \@bad displays the Bit allocation in Downstream
 * \param[in] argc number of arguments
 * \param[in] argv[0] 'bad'
 * \param[in] argv[1] 0=dec
 * \param[in] argv[1] 1=hex
 * \retval    0 error
 * \retval    1 success
 */
int dbgcmd_BitAllocDs (int argc, unsigned char **argv);


/**
 * This function called by \b \@bau displays the Bit allocation in Upstream
 * \param[in] argc number of arguments
 * \param[in] argv[0] 'bau'
 * \param[in] argv[1] 0=dec
 * \param[in] argv[1] 1=hex
 * \retval    0 error
 * \retval    1 success
 */
int dbgcmd_BitAllocUs (int argc, unsigned char **argv);

/**
 * This function called by \b \@bswp sets the bitwap ability n DS and/or US on or off
 * \param[in] argc number of arguments
 * \param[in] argv[0] 'bswp'
 * \param[in] argv[1] DS flag 0=BS off, 1=BS on
 * \param[in] argv[2] US flag 0=BS off. 1=BS on
 * \retval 0 error
 * \retval 1 success
 */
int dbgcmd_BitSwapOnOff (int argc, unsigned char **argv);

/** This function called by \b \@cd sets DSL to IDLE (deactivate)
 *  \param[in] argc number of arguments
 *  \param[in] argv[0] 'cd' show OS alarm status
 *  \retval 1 success (always)
 */
int dbgcmd_setDslIdle(int argc, unsigned char **argv);


/** This function called by \b \@ca sets DSL to ACTIVATE
 * \param[in] argc number of arguments
 * \param[in] argv[0] 'ca' show OS alarm status
 * \retval 1 success (always)
 */
int dbgcmd_setDslActivate(int argc, unsigned char **argv);


/** This function called by \b \@ce sets DSL to Retrain
 * \param[in] argc number of arguments
 * \param[in] argv[0] 'ce' show OS alarm status
 * \retval 1 success (always)
 */
int dbgcmd_setDslRetrain(int argc, unsigned char **argv);




/** Function called by \b \@dstones disables tones for training
 *  Tones can be either dec or hex, individual tones or range of tones.
 *  To specify the range of the tones, use MSB 01 for range start, MSB10 for range end'.
 *  You can use @bad after showtime to confirm the tones are disabled. (\sa dbgcmd_BitAllocDs)
 *  eg. @dstones 123 0x4080 0x8120 0x140 Will disable tone 123, 128 (0x80)-288 (0x120), and tone 320 (0x140).
 *  \param[in] argc number of arguments
 *  \param[in] argv[0] 'dstones'
 *  \param[in] argv[n] nth tone to be tuned
 *  \retval 0 error
 *  \retval 1 success
 */
int dbgcmd_dsTonesOnOff(int argc, unsigned char **argv);


/**
 *  Function called by \b \@pcb  shows, disables, enables PowerCutback
 *  \param[in] argc number of arguments
 *  \param[in] argv[0] 'pcb'
 *  \param[in] argv[1] optional: 0 disable
 *  \param[in] argv[1] optional: 0 enable
 *  \retval 0 error
 *  \retval 1 success
 */
int dbgcmd_setPowerCutBack(int argc, unsigned char **argv);


/** Function called by \b \@showtime displays showtime information
 *  \param[in] argv[0] 'showtime' dispay showtime information
 *  \retval 0 error
 *  \retval 1 success
 */
int dbgcmd_ShowTimeInfo(int argc, unsigned char **argv);


/** Function called by \b \@HLOGpsds shows Hlog downstream per subcarrier (\sa G997.1)
 *  \todo doxylink to G997.1
 *  \param[in] argc number of arguments
 *  \param[in] argv[0] 'HLOGpsds'
 *  \param[in] argv[1] 0 display decimal (default)
 *  \param[in] argv[1] 1 display hexadecimal
 *  \note only supported for ADSL2 and ADSL2+ mode
 *  \retval 0 error
 *  \retval 1 success
 */
int dbgcmd_HLOGpsds(int argc, unsigned char **argv);

// UR8_MERGE_START CQ11057 KCCHEN
int dbgcmd_pmdus(int argc, unsigned char **argv);
// UR8_MERGE_END CQ11057 KCCHEN

/** Function called by \b \@SNRpsds shows showtime SNR downstream per
 *  subcarrier (\sa G997.1)
 *  \param[in] argv[0] 'SNRpsds'
 *  \param[in] argv[1] 0 display decimal (default)
 *  \note only supported for ADSL2 and ADSL2+ mode
 *  \retval 0 error
 *  \retval 1 success
 */
int dbgcmd_SNRpsds(int argc, unsigned char **argv);


/** Function called by \b \@QLNpsds shows Quiet Line Noise per subcarrier
 *  \param[in] argc number of arguments
 *  \param[in] argv[0] 'QLNpsds'
 *  \param[in] argv[1] 0 display decimal (default)
 *  \param[in] argv[1] 1 display hexadecimal
 *  \note only supported for ADSL2 and ADSL2+ mode
 *  \retval 0 error
 *  \retval 1 success
 */
int dbgcmd_QLNpsds(int argc, unsigned char **argv);

//  UR8_MERGE_START CQ10979   Jack Zhang
/** Function called by \b \@HLINpsds shows linear channel characteristics per subcarrier
 *  \param[in] argc number of arguments
 *  \param[in] argv[0] 'HLINpsds'
 *  \param[in] argv[1] 0 display decimal (default)
 *  \param[in] argv[1] 1 display hexadecimal
 *  \note only supported for ADSL2 and ADSL2+ mode
 *  \retval 0 error
 *  \retval 1 success
 */
int dbgcmd_HLINpsds(int argc, unsigned char **argv);
//  UR8_MERGE_END   CQ10979*

//
//@quit : Quit the Dsl Debug App
//
int dbgcmd_quit(int argc, unsigned char **argv);


/// helper function to display bit allocation in downstream
void disp_bitalloc_ds(void);


/// helper function to display bit allocation in upstream
void disp_bitalloc_us(void);


/// helper function to display DSP error count
void disp_dsp_error(void);


/// helper function to display margin in downstream
void disp_margin_ds(void);

/** This function is invoked right after enter showtime.
 */
void display_ShowTimeInfo (tidsl_t *ptidsl);

/** Function called by \b \@eoc displays EOC information
 *  \param[in] argc number of arguments
 *  \param[in] argv[0] 'eoc' show EOC information (without additional parameters)
 *  \param[in] argv[N] data N to send
 *  \retval 1 success (always)
 */
int dbgcmd_EocInfo(int argc, unsigned char **argv);


/** Function called by \b \@mard dumps the downstream margin
 *  \param[in] argc number of parameters
 *  \param[in] argv[0] 'mard' dump the downstream margin
 *  \retval 0 error
 *  \retval 1 success
 */
int dbgcmd_Margin_Ds(int argc, unsigned char **argv);


/** Function called by \b \@dgasp sends dying gasp
 *  \param[in] argc number of parameters
 *  \param[in] argv[0] 'dgasp' display training parameters
 *  \retval 0 error
 *  \retval 1 success
 */
int dbgcmd_SendDyingGasp(int argc, unsigned char **argv);

/** This utility function is used to unify number given in dec or in hex
 *  \brief The function returns the number of digits (without leading [-]0x)
 *         given in string cp and converts the string cp, which can be dec or hex.
 *         The converted output is written to number, an previous allocated integer
 *  \param[in]  cp string to be converted
 *  \param[out] number pointer to converted integer output
 *  \return     number of digits of cp (eg. -0x244 -> 3)
 */
int GetNumber(unsigned char *cp,  int *number);


/** This utility function is used to convert hex value into dec value
 *  \param[in]  cp string to be converted (must not include 0x or numbers<0)
 *  \param[out] hex pointer to converted integer output
 *  \return     number of digits of cp (eg. 244 -> 3)
 */
int GetHexNumber (unsigned char *cp, unsigned int *hex);

char *AnnexFromNum (unsigned int annex_selected);

//----------------------------------------------

#endif  // 
