/****************************************************************************************
 *
 * FILE NAME:       efmcrc.c
 *
 * DESCRIPTION:     Linux EFM module CRC calculation header file.
 *
 * (C) Copyright 2007, Texas Instruments Inc
 *
 *
 * History:
 *   2/20/2007 CPH  CQ11466, Written, add EFM support
 *
 ****************************************************************************************/

#ifndef __EFMCRC_H__
#define __EFMCRC_H__

// ------------------ Define Compile Environment --------------------------------------
//
// Define EFMCRC_STANDALONE to generate standalone CRC testing program running
// under Windows
//
//#define EFMCRC_STANDALONE

#define ETH_TC_CRC16    // Handle US TC_CRC16 in the driver

enum {
  EFM_CRC_LITTLE_ENDIAN,  // Host is Little endian
  EFM_CRC_BIG_ENDIAN      // Host is Big endian 
};  

// EFM_CRC_ENDIAN setting is based on Host processor endianess.
// Change the following setting for your host processor
#define EFM_CRC_ENDIAN        EFM_CRC_LITTLE_ENDIAN

#ifndef EFMCRC_STANDALONE
#define EFMCRC_USE_INTERNAL_SDRAM
#endif


// ------------------------------------------------------------------------------------
// used by both CRC32 & CRC16 tables
#define CRCTBL_ENTRIES                  256

#define INTERNAL_SDRAM_START            0x80000000    // Start addr of internal SDRAM. (Total 4K)
#define CRC32_START_ADDR                (INTERNAL_SDRAM_START + 1024*3)   // last 1k of Internal SDRAM.
#define CRC16_START_ADDR                (CRC32_START_ADDR-512)   // CRC16 table size = 2*256 =512 bytes

// aka "CRC32-IEEE802.3"
// polynomial: X^32+X^26+X^23+X^22+X^16+X^12+X^11+X^10+X^8+X^7+X^5+X^4+X^2+X^1+X^0
#define CRC32_POLYNOMIAL                0xedb88320L


#ifdef ETH_TC_CRC16
//
// ETH_TC_CRC16 polynomial: x^16+x^12+x^5+1 (aka "CRC16-CCITT")
// As defined in 802.3-2005, 61.3.3.3
//
#define CRC16_POLYNOMIAL                0x8408

typedef union {
    unsigned char   b[2];
    unsigned short  x;
} ETH_CRC16;

#endif


typedef union {
    unsigned char   b[4];
    unsigned int    x;
} ETH_CRC32;


// ---------------- Begin Prototype -------------------------

void efmcrc_init_crcTable(void);
unsigned long efm_calc_crc32(unsigned char *buf, int len);
unsigned short efm_calc_crc16(unsigned char *buf, int len);
void efmcrc_load_pdsp_crc16_lut(void);

// ---------------- End Prototype  --------------------------

#endif  // __EFMCRC_H__
