/****************************************************************************************
 *
 * FILE NAME:       tn7efm.c
 *
 * DESCRIPTION:     Linux EFM module high level function.
 *
 * (C) Copyright 2007, Texas Instruments Inc
 *
 * History:
 *   2/20/2007 CPH  CQ11466 Written, add EFM support.
 *   9/10/2007 CPH  CQ11466 Added EFM/ATM auto-switch support.
 *   2/16/2008 MK   SMS00854598: Fixed PTM routed mode.
 *  10/20/2009 EYin SMS01234914: Added PTM-preemption feature.
 ****************************************************************************************/
 
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/atmdev.h>
#include <linux/delay.h>
#include <linux/spinlock.h>
#include <linux/smp_lock.h>
#include <asm/io.h>
#include <asm/mips-boards/prom.h>
#include <linux/proc_fs.h>
#include <linux/string.h>
#include <linux/ctype.h>
#include <linux/etherdevice.h> /* eth_type_trans */

#include <linux/skbuff.h>

#include "dsl_hal_api.h"
#include "dsl_hal_support.h"
#include "dev_host_interface.h"

#define _CPHAL_AAL5
#define _CPHAL_SAR
#define _CPHAL_HAL
typedef void OS_PRIVATE;
typedef void OS_DEVICE;
typedef void OS_SENDINFO;
typedef void OS_RECEIVEINFO;
typedef void OS_SETUP;

#include "tn7efm.h"
#include "cpswhal_cpsar.h"
#include "tn7atm.h"
#include "tn7api.h"
#include "version.h"
#define _SIZE_T_
#include "dsl_hal_register.h"

#include "tn7efm_pri.h"
#include "efmcrc.h"

extern unsigned int def_sar_inter_pace;
extern void tn7sar_switchmode_teardownATM(struct atm_dev *dev);
extern int tn7atm_irq_request (struct atm_dev *dev);
extern void tn7atm_close (struct atm_vcc *vcc);
extern void *tn7dsl_get_pIhw(void);
extern void tn7sar_SetEFMmode(Tn7AtmPrivate *priv, int mode);
extern void tn7sar_SetOSDev(Tn7AtmPrivate *priv, void *os_dev);


#ifdef EFM_DEBUG
extern int __guDbgLevel;
extern unsigned int g_dsl_current_sync;
extern unsigned int g_dsl_unsync_count;
#endif

extern struct atm_dev *mydev;

EFM_STATS efm_stats;

unsigned int g_efm_proc_ctl = 0;

//static unsigned int efm_tx_timeout_val = AR7_EFM_DEFAULT_TIMEOUT;
struct net_device *mydev_efm;
Tn7AtmPrivate *mypriv;
int toggle;

#define PACKET_PRIORITY 0x3 //bit0 and bit1 
#define BAND0_HIGH_PRIORITY 0x01 //band0 high priority packets

int tn7efm_modeSwitch_closeEFM(void)
{
  Tn7AtmPrivate *priv = mypriv;
  HAL_FUNCTIONS *pHalFunc;
  HAL_DEVICE    *pHalDev;
  int Ret;

#ifdef EFM_DEBUG
  printk("tn7efm_modeSwitch_closeEFM\n");
#endif

  dgprintf(4, "tn7efm_modeSwitch_closeEFM()\n");
  pHalFunc = (HAL_FUNCTIONS *)priv->pSarHalFunc;
  pHalDev  = (HAL_DEVICE *)priv->pSarHalDev;

  if ((Ret = pHalFunc->Close(pHalDev, 2)) != 0) //2: Full teardown
    return Ret;

  memset (&priv->lut[RESERVED_OAM_CHANNEL], 0, sizeof (priv->lut[RESERVED_OAM_CHANNEL]));

  if (priv->sar_irq)
  {
    free_irq (priv->sar_irq, priv->efm_dev);  // teardown sar irq
    priv->sar_irq = 0; // to indicate the SAR irq is gone
  }

  return 0;
}

int tn7efm_get_currTCmode(void)
{
  Tn7AtmPrivate *priv = mypriv;
  return priv->curr_TC_mode;
}

int tn7efm_modeSwitch_closeATM(void)
{
  Tn7AtmPrivate *priv = mypriv;
  HAL_FUNCTIONS *pHalFunc;
  HAL_DEVICE    *pHalDev;
  int Ret;

#ifdef EFM_DEBUG
  printk("tn7efm_modeSwitch_closeATM\n");
#endif

  dgprintf(4, "tn7efm_modeSwitch_closeATM()\n");
  pHalFunc = (HAL_FUNCTIONS *)priv->pSarHalFunc;
  pHalDev  = (HAL_DEVICE *)priv->pSarHalDev;

  if ((Ret = pHalFunc->Close(pHalDev, 2)) !=0)     // 2: Full Teardown
    return Ret;

  memset (&priv->lut[RESERVED_OAM_CHANNEL], 0, sizeof (priv->lut[RESERVED_OAM_CHANNEL]));

 /*
  * freeup irq's
  */
  if (priv->sar_irq)
  {
    free_irq (priv->sar_irq, priv->dev);  // teardown sar irq
    priv->sar_irq = 0; // to indicate the SAR irq is gone
  }

  return 0;
}

void tn7efm_modeSwitch_teardownATM(void)
{
  Tn7AtmPrivate *priv = mypriv;
  int chan;

  dgprintf(4, "tn7efm_modeSwitch_teardownATM()\n");
#ifdef EFM_DEBUG
  printk("tn7efm_modeSwitch_teardownATM\n");
#endif

  for (chan=0; chan<MAX_DMA_CHAN; chan++)
  {
    if (priv->lut[chan].ready)
    {
#ifdef EFM_DEBUG
      printk("ATM closing chan%d\n", chan);
#endif
      tn7atm_close (priv->lut[chan].vcc);
    }
  }

  tn7sar_deactivate_vc(priv, RESERVED_OAM_CHANNEL);//15);  de-activate oam channel
}

void tn7efm_modeSwitch_teardownEFM(void)
{
  Tn7AtmPrivate *priv = mypriv;
  int chan;

  dgprintf(4, "tn7efm_modeSwitch_teardownEFM()\n");

#ifdef  EFM_DEBUG
  printk("tn7efm_modeSwitch_teardownEFM\n");
#endif

  // send netlink socket to upper layer

  for (chan=0; chan<NUM_EFM_CHAN ; chan++)
    tn7sar_deactivate_vc(priv, chan);

  tn7efm_dev_close(priv->efm_dev);
}

void tn7efm_modeSwitch_initATM(void)
{
  Tn7AtmPrivate *priv = mypriv;

  dgprintf(4, "tn7efm_modeSwitch_initATM()\n");
#ifdef EFM_DEBUG
  printk("tn7efm_modeSwitch_initATM\n");
#endif

  priv->curr_TC_mode = TC_MODE_ATM;
  tn7sar_SetEFMmode (priv, 0);  // EFM mode is the lsb 8 bit of EFM_CTL
  tn7sar_SetOSDev (priv, priv->dev);

  if (tn7sar_init (mydev, priv) != 0)
  {
    printk ("Error: Failed to switch to ATM mode!\n");
  }

  memset (&priv->stats, 0, sizeof(struct net_device_stats));

  tn7atm_irq_request(mydev);
}

void tn7efm_modeSwitch_initEFM(void)
{
  dgprintf(4, "tn7efm_modeSwitch_initEFM()\n");
#ifdef EFM_DEBUG
  printk("tn7efm_modeSwitch_initEFM\n");
#endif

  if (tn7efm_detect ())
  {
    printk ("Error : Failed to Initialize EFM driver !!\n");
    return;
  }
}


static int tn7efm_sar_handle_interrupt(Tn7AtmPrivate *priv, int *more)
{
  int rc=0;
  int (*halIsr)(HAL_DEVICE *halDev, int *work);

  halIsr = priv->halIsr;

  rc = halIsr((HAL_DEVICE *)priv->pSarHalDev, more);

  /* Do EOI if the interrupt was from a valid source. */
  if(rc != EC_VAL_INTERRUPT_NOT_FOUND)
  {
        ((HAL_FUNCTIONS *)priv->pSarHalFunc)->PacketProcessEnd((HAL_DEVICE *)priv->pSarHalDev);
  }

  return rc;
}


//irqreturn_t DDA_efm_sar_irq (int irq, void *voiddev, struct pt_regs *regs)
irqreturn_t tn7efm_sar_irq (int irq, void *voiddev, struct pt_regs *regs)
{
  struct net_device *dev;
  Tn7AtmPrivate *priv;
  int more;

#ifdef TIATM_INST_SUPP
  psp_trace (ATM_DRV_SAR_ISR_ENTER);
#endif

  dev = (struct net_device *) voiddev;
  priv = (Tn7AtmPrivate *) dev->priv;

#ifdef EFM_DEBUG
  if (priv != mypriv)
  {
    printk("err:tn7efm_sar_irq, priv=%08x, expect=%08x\n",
      (unsigned int) priv, (unsigned int) mypriv);
  }
#endif

  tn7efm_sar_handle_interrupt (priv, &more);

  dgprintf (6, "Leaving DDA_atm_sar_irq\n");
}

void tn7efm_initenv (void)
{
  Tn7AtmPrivate *priv=mypriv;
  char *penv;
  char *ptr;
  unsigned int efm_mac[6];

  if ((penv = prom_getenv("EFM_CTL")) != NULL)
  {
    priv->efm_ctl = os_atoih(penv);
  }
  else {
    priv->efm_ctl = EFM_LOAD_NORM;
  }

  if ( (penv = prom_getenv("EFM_PROC_CTL")) != NULL)
  {
    g_efm_proc_ctl = os_atoih(penv);
  }

  memcpy (priv->efm_mac_addr, EFM_DEFAULT_MAC, ETH_ALEN);
  if ( (penv = prom_getenv("EFM_MAC")) != NULL)
  {
    if (ETH_ALEN==sscanf(penv,"%02X:%02X:%02X:%02X:%02X:%02X",
        (unsigned int *)&efm_mac[0], (unsigned int *) &efm_mac[1],
        (unsigned int *)&efm_mac[2], (unsigned int *) &efm_mac[3],
        (unsigned int *)&efm_mac[4], (unsigned int *) &efm_mac[5]) )
    {
      int i;
      for (i=0; i<ETH_ALEN; i++)
        priv->efm_mac_addr[i] = (unsigned char) (efm_mac[i] & 0xFF);
    }
  }

#ifdef EFM_DEBUG
  penv = prom_getenv("DbgLevel");
  if (penv)
  {
    __guDbgLevel = os_atoi(penv);
  }
#endif

  priv->efm_sarRxBuf = EFM_RX_BUFFER_NUM_DEFAULT;
  ptr = prom_getenv ("EFM_SarRxBuf");
  if (ptr)
  {
    priv->efm_sarRxBuf = os_atoi (ptr);
  }

  priv->efm_sarRxMax = EFM_RX_SERVICE_MAX_DEFAULT;
  ptr = prom_getenv ("EFM_SarRxMax");
  if (ptr)
  {
    priv->efm_sarRxMax = os_atoi (ptr);
  }

  priv->efm_sarTxBuf = EFM_TX_BUFFER_NUM_DEFAULT;
  ptr = prom_getenv ("EFM_SarTxBuf");
  if (ptr)
  {
    priv->efm_sarTxBuf = os_atoi (ptr);
  }

  priv->efm_sarTxMax = EFM_TX_SERVICE_MAX_DEFAULT;
  ptr = prom_getenv ("EFM_SarTxMax");
  if (ptr)
  {
    priv->efm_sarTxMax = os_atoi (ptr);
  }

#ifdef EFM_DEBUG
  printk("EFM_CTL =%08x, EFM_PROC_CTL=%08x\n", priv->efm_ctl, g_efm_proc_ctl);
#endif

}


static int tn7efm_irq_request (Tn7AtmPrivate * priv)
{
  dgprintf (4, "tn7atm_irq_request()\n");

  /*
   * Register SAR interrupt
   */
  priv->sar_irq = LNXINTNUM (ATM_SAR_INT);      /* Interrupt line # */

  if (request_irq (priv->sar_irq, tn7efm_sar_irq, SA_INTERRUPT, "SAR ", priv->efm_dev))
    printk ("Could not register tn7atm_sar_irq\n");

  /*
   * interrupt pacing
   */
  avalanche_request_pacing (priv->sar_irq, ATM_SAR_INT_PACING_BLOCK_NUM,
                            def_sar_inter_pace);

  return 0;
}

enum {
  TCMODE_SWITCH_TEARDOWN,
  TCMODE_SWITCH_WAIT_TEARDOWN,
  TCMODE_SWITCH_CLOSE,
  TCMODE_SWITCH_SETNEWMODE,
  TCMODE_SWITCH_DONE,
};

void tn7efm_handle_autoswitch_tasklet (unsigned long data)
{
extern int efm_IsTearDownPending(HAL_DEVICE *HalDev, int num_chan);
  Tn7AtmPrivate *priv=mypriv;
  int state= TCMODE_SWITCH_TEARDOWN;
  int target_TC_mode = priv->target_TC_mode;
  int num_chan = (priv->curr_TC_mode == TC_MODE_PTM ? NUM_EFM_CHAN : 16);
  int Ret;

#ifdef EFM_DEBUG
  printk("tn7efm_handle_autoswitch_tasklet\n");
#endif

  if (priv->curr_TC_mode == target_TC_mode)  // current mode is same as the target mode
    return;  // do nothing

  priv->mode_switching=1;  // start mode switching

  while (state != TCMODE_SWITCH_DONE)
  {
    switch (state) {
    case  TCMODE_SWITCH_TEARDOWN:
      if (target_TC_mode== TC_MODE_PTM)
            tn7efm_modeSwitch_teardownATM();
      else  tn7efm_modeSwitch_teardownEFM();
      state = TCMODE_SWITCH_WAIT_TEARDOWN;  // move on to next state
      break;

    case TCMODE_SWITCH_WAIT_TEARDOWN:
      // It is important to make sure we receive confirmation from pdsp that
      // the teardown is complete (PDSP will generate TxTeardown/RxTeardown interrupt)
      // before we proceed to close the current TC mode.
      if (efm_IsTearDownPending(priv->pSarHalDev, num_chan))
           break;
      state = TCMODE_SWITCH_CLOSE;  // move on to next state
      break;

    case TCMODE_SWITCH_CLOSE:
      if (target_TC_mode== TC_MODE_PTM)
            Ret = tn7efm_modeSwitch_closeATM();
      else  Ret = tn7efm_modeSwitch_closeEFM();
      state = TCMODE_SWITCH_SETNEWMODE;   // move on to next state
      break;

    case TCMODE_SWITCH_SETNEWMODE:
      if (target_TC_mode== TC_MODE_PTM)
            tn7efm_modeSwitch_initEFM();      // Setup EFM mode
      else  tn7efm_modeSwitch_initATM();      // setup ATM mode
      state = TCMODE_SWITCH_DONE;
      priv->mode_switching=0; // mode switching end
      break;
    } // switch
  } // while

}

int tn7efm_init (struct net_device *dev)
{
  Tn7AtmPrivate *priv= (Tn7AtmPrivate *) netdev_priv(dev);

  if (tn7efm_sar_init (dev) != 0)
  {
    printk ("Failed to init EFMSAR.\n");
    return -ENODEV;
  }

  priv->pIhw = tn7dsl_get_pIhw();

  //Now that all has been initialized.
  priv->efm_dev =dev;

  return 0;
}


int tn7efm_register (Tn7AtmPrivate * priv)
{
  /*
   * allocate memory for the device
   */
  struct net_device *dev;
  int ret;

  dgprintf(4, "tn7efm_register\n");

  dev = kmalloc(sizeof(struct net_device), GFP_KERNEL);
  mydev_efm = dev; // save it
  memset(dev, 0, sizeof(struct net_device));

  ether_setup(dev); /* assign some of the fields */

  dev->open            = tn7efm_dev_open;
  dev->stop            = tn7efm_dev_close; // tn7efm_release;
  dev->set_config      = tn7efm_config;
  dev->hard_start_xmit = tn7efm_send;
  dev->do_ioctl        = tn7efm_ioctl;
  dev->get_stats       = tn7efm_stats;
  dev->change_mtu      = tn7efm_change_mtu;
//  dev->rebuild_header  = tn7efm_rebuild_header;
//  dev->hard_header     = tn7efm_header;
  dev->set_mac_address = tn7efm_set_mac_addr;

#if 0 //cph99 disable watcdog for now...
  dev->tx_timeout      = tn7efm_tx_timeout;
  dev->watchdog_timeo  = efm_tx_timeout_val; // set default timeout value
#endif

  /* keep the default flags, just add NOARP */
//  dev->flags           |= IFF_NOARP;
  dev->features        |= NETIF_F_NO_CSUM;
  dev->hard_header_cache = NULL;      /* Disable caching */
  priv->efm_name = "TI Avalanche EFM SAR";
  printk ("registered device %s\n", priv->efm_name);

  dev->priv = priv;       /* setup device data in atm_dev struct */
  priv->efm_dev = dev;            /* setup atm_device in avalanche sar struct */

  strcpy (dev->name, "ptm%d");

  if ((ret = register_netdev(dev)))
    printk("tn7efm: error %i registering device \"%s\"\n", ret , dev->name);
  else
    ret = 0;

  return EFM_REG_OK;
}


int tn7efm_detect (void)
{
  Tn7AtmPrivate *priv= mypriv;
  int efm_mode;

  dgprintf (4, "tn7efm_detect().\n");

  priv->curr_TC_mode = TC_MODE_PTM;

  efm_mode = priv->efm_ctl & 0xFF;  // EFM mode is the lsb 8 bit of EFM_CTL
  tn7sar_SetEFMmode (priv, efm_mode);
  tn7sar_SetOSDev (priv, priv->efm_dev);

  memset (&priv->stats, 0, sizeof (struct net_device_stats));

  if (!priv->efm_dev)
  {
    if ((tn7efm_register(priv)) == EFM_REG_FAILED)
    return -ENODEV;
  }

  if (tn7efm_init (priv->efm_dev)) // priv->efm_dev was set in tn7efm_register()
  {
    printk ("Error : Failed to Initialize the DSL subsystem !!.\n");
    return -ENODEV;
  }

  if (tn7efm_irq_request (priv) < 0)
    return -EBUSY;

  priv->efm_initialized=1;

  return 0;
}


/*
 * tn7efm_dev_open:
 *   Note: For EFM, all channels (prep, non-prep) are opened in device open here.
 *         This is different from ATM case.
 */
static int tn7efm_dev_open (struct net_device *dev)
{
  Tn7AtmPrivate *priv = netdev_priv(dev);
  int chan;
  int rc;

  dgprintf (1, "Enter tn7efm_dev_open\n");
#ifdef EFM_DEBUG
  printk("tn7efm_dev_open\n");
#endif

  /* The first byte is '\0' to avoid being a multicast
   * address (the first byte of multicast addrs is odd).
   */
  memcpy(dev->dev_addr, priv->efm_mac_addr, ETH_ALEN);

  if(!priv)
  {
    printk("tn7efm_dev_open:null priv\n");
    return -1;
  }

  for (chan=0; chan<NUM_EFM_CHAN; chan++)
    tn7efm_set_lut(dev, chan);

   /* These are used for the Input queue throttling. */
  for (chan=0; chan<NUM_EFM_CHAN; chan++)
    priv->available_tx_bufs[chan] = priv->sarTxBuf;

   /* Activate SAR channel */
   for (chan=0; chan<NUM_EFM_CHAN; chan++)
   {
      rc = tn7efm_sar_activate_chan (dev, chan);
      if (rc < 0)
      {
        printk("tn7efm_dev_open:failed to activate hw channel\n");
        for (chan=0; chan<NUM_EFM_CHAN; chan++)
          tn7efm_lut_clear(dev, chan);
        return -EBUSY;
      }
    }

  for (chan=0; chan<NUM_EFM_CHAN; chan++)
    priv->lut[chan].ready = 1;

  netif_start_queue(dev);

  dgprintf (1, "Leave tn7efm_dev_open\n");
  return 0;
}

static int tn7efm_dev_close (struct net_device *dev)
{
  Tn7AtmPrivate *priv = netdev_priv(dev);
  int chan;
  spinlock_t closeLock;
  unsigned int closeFlag;

  dgprintf(4, "tn7efm_dev_close\n");

#ifdef EFM_DEBUG
  printk("tn7efm_dev_close\n");
#endif

  spin_lock_irqsave (&closeLock, closeFlag);

  for (chan=0; chan<NUM_EFM_CHAN; chan++)
  {
    priv->lut[chan].bClosing = 1;
    priv->lut[chan].ready = 0;
    priv->available_tx_bufs[chan] = 0;
    tn7efm_lut_clear(dev, chan);
  }

  spin_unlock_irqrestore (&closeLock, closeFlag);
  dgprintf(4, "tn7efm_dev_close done\n");
  return 0;
}

static int tn7efm_config(struct net_device *dev, struct ifmap *map)
{
  dgprintf(4, "tn7efm_config\n");

  if (dev->flags & IFF_UP) /* can't act on a running interface */
          return -EBUSY;

  /* Don't allow changing the I/O address */
  if (map->base_addr != dev->base_addr) {
    printk(KERN_WARNING "tn7efm_config: Can't change I/O address\n");
    return -EOPNOTSUPP;
  }

  /* Allow changing the IRQ */
  if (map->irq != dev->irq) {
          dev->irq = map->irq;
          /* request_irq() is delayed to open-time */
  }

  /* ignore other fields */
  return 0;
}

static int tn7efm_set_mac_addr(struct net_device *p_dev,void * addr)
{
  struct sockaddr *sa = addr;

#ifdef EFM_DEBUG
  Tn7AtmPrivate *priv = netdev_priv(p_dev);
  int i;

printk("tn7efm_set_mac_addr\n");
printk("sa_data=\n");
for(i=0; i<14; i++)
  printk("%02x ", sa->sa_data[i]);
printk("\n");

printk("curr dev_addr:\n");
for(i=0; i<6; i++)
  printk("%02x ", p_dev->dev_addr[i]);
printk("\n");
#endif


//    AddressParams.macAddress = sa->sa_data;


//    memcpy(hDDA->macAddr,sa->sa_data,p_dev->addr_len);
    memcpy(p_dev->dev_addr,sa->sa_data,p_dev->addr_len);
    return 0;
}


static int tn7efm_send(struct sk_buff *skb, struct net_device *dev)
{
  Tn7AtmPrivate *priv = netdev_priv(dev);
  PITIDSLHW_T pIhw = priv->pIhw;
  int priority = 0; // hard code to use priority 0 (hi-pri queue) for EFM
  int chan;
  int bret;

  dgprintf(4, "tn7efm_send\n");

  efm_stats.send_count++;

#ifdef EFM_DEBUG
  if (g_efm_proc_ctl & 1)
  {
    printk("\nskb->len=%d, data_len=%d, truesize=%d, sizeof(skbuf)=%d, skb->data=%08x\n",
      skb->len, skb->data_len,
      skb->truesize, sizeof(struct sk_buff),
      (unsigned int) skb->data);

  if (skb->data)
    xxdump(skb->data, skb->len, "efmTx_data:");
  }
#endif

  if ((priv->lConnected != 1) || (priv->xmitStop == 1)) // don't send if DSL not show time
  {
#ifdef EFM_DEBUG
    if (!(g_efm_proc_ctl & 0x10) ) // so that loopback pkt won't be dropped
#endif
    {
      dev_kfree_skb (skb);
      priv->stats.tx_errors++; //UR8_MERGE_START_END CQ10700 Manjula K
      return 0; // you have to return 0, or the upper layer will keep trying, eventually crash...
    }
  }

  //if receive high priority packet and PTM US pre-emption is supported 
  if ( ((skb->cb[47] & PACKET_PRIORITY) == BAND0_HIGH_PRIORITY) &&
    //check to see CO support US pre-emption or not
       (pIhw->PTM_pre_emption & PTM_PREEMPT_US_ENABLED) )
  {    
     chan = 1;  //hi-pri data from band0
     priority =1;
  }
  else
  {     
    chan=0;
    priority=0;          
  }  

  if (priv->lut[chan].bClosing == 1)
  {
    netif_stop_queue(skb->dev);
    dev_kfree_skb(skb);
    priv->stats.tx_errors++; //UR8_MERGE_START_END CQ10700 Manjula K
    return 1;
  }

  //bret = tn7efm_queue_packet_to_sar (dev, skb, chan);

  bret = tn7efm_queue_packet_to_sar (dev, skb, chan, priority);
  
  dgprintf(4, "tn7efm_send done\n");
//  return bret;
  return 0;
}


inline static int tn7efm_queue_packet_to_sar (void *dev1, void *skb1, int dmachan,int priority)
{
  struct sk_buff *skb = (struct sk_buff *) skb1;
  //int priority = 0; // hard code to use priority 0 (hi-pri queue) for EFM
  struct net_device *dev= (struct net_device *) dev1;
  Tn7AtmPrivate *priv = (Tn7AtmPrivate *) netdev_priv(dev);
  int retval;
  int flag = 0;
  int tailroom;
  unsigned char *ptail;
  unsigned int padding;
  ETH_CRC32 crc32;
#ifdef ETH_TC_CRC16
  ETH_CRC16 crc16;
#endif
  int i;
  int tx_pkt_len;

  dgprintf(4, "tn7efm_queue_packet_to_sar\n");

  efm_stats.queue_packet_to_sar_count++;
 

  /*
   * PANKAJ: Once a packet has been passed down to the HAL layers we need to
   * request the upper layers to stop pushing data to us. This is to ensure
   * that higher priority packets are not being held back by the lower
   * priority ones.
   */
    
   if (skb->dev != NULL)
   {
    atomic_dec ((atomic_t *) &priv->available_tx_bufs[dmachan]);
    if(priv->available_tx_bufs[dmachan]<= 0)
    {
                
        netif_stop_queue(skb->dev);
       
      flag = 1;
    }
  }

// add crc32 here...
  if (skb->len<60)
  {
    padding = 60-skb->len;
  }
  else padding = 0;

#ifdef EFM_DEBUG
  if (g_efm_proc_ctl & 4)
     printk("padding=%d, tailroom=%d\n", padding, skb_tailroom(skb));
#endif

  tx_pkt_len = skb->len; // save the org skb length before add padding & FCS.


  if ((tailroom = skb_tailroom(skb)) < (padding+EFM_TRAILOR_LEN)) // reserve 4(+2)bytes for CRC32 (&TC_CRC16)
  {
    struct sk_buff *skb_tmp= skb;

#ifdef EFM_DEBUG
    //[cph] You don't want to do this too often...
    if (g_efm_proc_ctl & 4)
      printk("skb_copy_expand: skb tailroom (%d) < padding+%d (%d)\n", tailroom, EFM_TRAILOR_LEN, padding+EFM_TRAILOR_LEN);
#endif

    skb= skb_copy_expand(skb_tmp, 0, padding+EFM_TRAILOR_LEN, GFP_ATOMIC);
    dev_kfree_skb(skb_tmp);

    if (skb==NULL)
    {
#ifdef EFM_DEBUG
      printk("skb_copy_expand failed!\n");
#endif
      priv->stats.tx_errors++;
      priv->stats.tx_dropped++;
      return 0;
    }
  }

#ifdef EFM_DEBUG
  if (g_efm_proc_ctl & 4)
  {
    printk("\nskb->len=%u, data_len=%d, truesize=%d, sizeof(skbuf)=%d, skb->data=%08x\n",
      skb->len, skb->data_len,
      skb->truesize, sizeof(struct sk_buff),
      (unsigned int) skb->data);
  }
#endif

  ptail = (unsigned char *) skb_put(skb, padding+EFM_TRAILOR_LEN); // extend skb data by 'padding+FCS(4)+CRC16(2)'

  for(i=0; i<padding; i++)
    *ptail++= 0;  // add padding if org packet size < 60

  crc32.x = efm_calc_crc32(skb->data, skb->len-EFM_TRAILOR_LEN); // don't count FCS & CRC16 for FCS

#if (EFM_CRC_ENDIAN==EFM_CRC_LITTLE_ENDIAN)
  // Host is Little Endian
  *ptail++ = crc32.b[0];
  *ptail++ = crc32.b[1];
  *ptail++ = crc32.b[2];
  *ptail++ = crc32.b[3];

#else
  // Host is Big Endian
  *ptail++ = crc32.b[3];
  *ptail++ = crc32.b[2];
  *ptail++ = crc32.b[1];
  *ptail++ = crc32.b[0];
#endif



#ifdef ETH_TC_CRC16
#ifdef EFM_DEBUG
  if (g_efm_proc_ctl & 0x80)
  {
    crc16.x = 0xAABB;
  }
  else
#endif
  {
   crc16.x = efm_calc_crc16(skb->data, skb->len-TC_CRC16_LEN);
  }

#if (EFM_CRC_ENDIAN==EFM_CRC_LITTLE_ENDIAN)
  *ptail++ = crc16.b[0];
  *ptail = crc16.b[1];
#else
  // Big Endian
  *ptail++ = crc16.b[1];
  *ptail = crc16.b[0];
#endif

#endif

#ifdef EFM_DEBUG
  if (g_efm_proc_ctl & 4)
    printk("new skb->len=%d, EFM_CRC32:%08x, EFM_CRC16:%04x\n",  skb->len, crc32.x, crc16.x);
#endif


  if ((retval = tn7efm_sar_send_packet (dev, dmachan, skb, skb->data, skb->len, priority)))
  {
      dev_kfree_skb(skb);

    /*
     * PANKAJ: In case of error we need to wakeup the queue again.
     */
        //AV_26 Probably not needed for now
    //if ((skb->dev != NULL) /*&& (EnableQoS)*/)
    if ((skb->dev != NULL) && (flag))
      {
        netif_wake_queue(skb->dev);
        flag = 0;
      }
    priv->stats.tx_dropped++;
  }
  else
  {
    // Note:
    //  We can't do tx_bytes in tn7efm_send_complete() as the skb->len
    //  in that function might include extra bytes for paddings & not representing
    //  the original data that was passed down from upper layer.
    priv->stats.tx_bytes += tx_pkt_len;
  }

  dgprintf(4, "tn7efm_queue_packet_to_sar done\n");
  return retval;
}

static int tn7efm_sar_send_packet(struct net_device *dev, int chan, void *new_skb, void *data,unsigned int len, int priority)
{
  Tn7AtmPrivate *priv = (Tn7AtmPrivate *) netdev_priv(dev);
  FRAGLIST fragList;
  unsigned int mode;
  OS_FUNCTIONS *os_funcs;
  HAL_FUNCTIONS *pHalFunc;
  HAL_DEVICE    *pHalDev;
  int retval;

  dgprintf(4, "tn7efm_sar_send_packet, priority=%d\n", priority);

#ifdef EFM_DEBUG
  if (g_efm_proc_ctl & 4)
    printk("tn7efm_sar_send_packet\n" );
#endif

  os_funcs= (OS_FUNCTIONS *) priv->pSarOsFunc;
  pHalFunc = (HAL_FUNCTIONS *)priv->pSarHalFunc;
  pHalDev  = (HAL_DEVICE *)priv->pSarHalDev;

  fragList.len = len;
  fragList.data = (void *)data;


#ifdef EFM_DEBUG
    if (g_efm_proc_ctl & 1)
    {
        xxdump((char *)fragList.data , fragList.len, "efmTx_pad:");
    }
#endif

  /*mode bit
  31-19 unused
  18    oam cell, 1 = true, 0=false
  17-16 oam type, 0=F4 seg, 1=F4 End, 2=F5 seg, 3=F5 end
  15-08 transimit queue, current, 0=priority queue, 1=normal queue
  07-00 channel number
  */
  mode = 0;
  mode |= (0xff & chan);
  mode |= ((0xff & priority) << 8);
  mode |= (PACKET_TYPE_NULL <<16);

  dgprintf(4, "mode = 0x%x\n", mode);
  
  os_funcs->DataCacheHitWriteback(fragList.data, len);
  if((retval = pHalFunc->Send(pHalDev, &fragList, 1, len, new_skb, mode)) != 0)
  {
    dgprintf(1, "EFMSAR hal failed to send packet.\n");
    return retval;
  }

#ifndef AR7_EFM
  sarStat.txPktCnt++;
  sarStat.txBytes +=len;
#endif

  dgprintf(4, "tn7efm_sar_send_packet done\n");
  return 0;
}

int tn7efm_send_complete (void *skb1)
{
  Tn7AtmPrivate *priv=mypriv;
  PITIDSLHW_T pIhw = priv->pIhw;
  struct sk_buff *skb = (struct sk_buff *) skb1;
  static unsigned int ledticks;
  int chan=0;
  struct net_device *skb_dev=NULL;

  dgprintf (4, "tn7efm_send_complete\n");

#ifdef TIATM_INST_SUPP
  psp_trace (ATM_DRV_PKT_TX_COMPLETE_ENTER);
#endif

  if (skb)
  {
      priv->stats.tx_packets++;
      
      if ( ((skb->cb[47] & PACKET_PRIORITY) == BAND0_HIGH_PRIORITY) &&
            (pIhw->PTM_pre_emption & PTM_PREEMPT_US_ENABLED) )
        chan=1;
      else 
        chan=0;        

      skb_dev = skb->dev;
      // Note:
      //  We can't do tx_bytes here as the skb->len here might include
      //  padding which is not what we want.
      // priv->stats.tx_bytes += skb->len;
      dev_kfree_skb(skb);
  }

  if (priv->lut[chan].bClosing != 1)
  {
    atomic_inc((atomic_t *) &priv->available_tx_bufs[chan]);
  }

//  atomic_inc (&vcc->stats->tx);

//AV_26
#if defined (CONFIG_LED_MODULE) || defined (CONFIG_MIPS_AVALANCHE_COLORED_LED)
  if ((ledticks != jiffies) && (hnd_LED_0))
  {
    ledticks = jiffies;
    TN7DSL_LED_ACTION(hnd_LED_0, MOD_ADSL, DEF_ADSL_ACTIVITY);
  }
#endif

  /*
   * track number of buffer used
   */
  if (skb_dev != NULL)
  {
    netif_wake_queue (skb_dev);
  }

#ifdef TIATM_INST_SUPP
  psp_trace_par (ATM_DRV_PKT_TX_COMPLETE_EXIT, skb->len);
#endif

  dgprintf (4, "tn7efm_send_complete() done\n");
  return 0;
}

int tn7efm_sar_receive(OS_DEVICE *os_dev, FRAGLIST *frag_list, unsigned int frag_count, unsigned int packet_size,
                 HAL_RECEIVEINFO *hal_receive_info, unsigned int mode)
{
  int ch;
//  Tn7AtmPrivate *priv = (Tn7AtmPrivate *) os_dev;
  Tn7AtmPrivate *priv = (Tn7AtmPrivate *) mypriv;
  HAL_FUNCTIONS *pHalFunc;
  int bRet;
  static int cnt = 0;

  dgprintf(4, "tn7efm_sar_receive\n");

#ifdef EFM_DEBUG
  if (priv != mypriv)
  {
    printk("err: tn7efm_sar_receive, priv=%08x, expect:%08x\n",
      (unsigned int) priv, (unsigned int) mypriv);
  }
#endif

  #ifdef TIATM_INST_SUPP
  psp_trace(ATM_DRV_RX_ENTER);
  #endif

  pHalFunc = (HAL_FUNCTIONS *)priv->pSarHalFunc;

  /* Mode contains channel info */
  ch = (mode & 0xFF);

#ifdef EFM_DEBUG
  if (ch!=0)
  {
    printk("err: tn7efm_sar_receive ch not zero (%d)\n", ch);
  }
#endif

  /*** Should we remove this check when we support multiple fragments in the Fraglist. ***/
  if(frag_count > 1 || frag_list->len == 0)
  {
    printk ("Packet fragment count > 1, not handle.\n");
    bRet = pHalFunc->RxReturn(hal_receive_info, 0);
    #ifdef TIATM_INST_SUPP
    psp_trace_par(ATM_DRV_RX_EXIT, bRet);
    #endif

    return 1;
  }

  /* Added check to see if the channel is closed */
  if(!priv->lut[ch].ready)
  {
    if((cnt++ %100) == 0)
        printk ("Channel %d is closed. cnt=%d\n", ch, cnt);

    pHalFunc->RxReturn(hal_receive_info, 0);
    return 1;
  }

  tn7atm_data_invalidate (frag_list->data, (int)frag_list->len);
  bRet = tn7efm_receive(priv->efm_dev, ch, packet_size, frag_list->OsInfo, frag_list->data);

  if(bRet==0)
  {
#ifndef AR7_EFM
    sarStat.rxPktCnt++;
    sarStat.rxBytes += packet_size;
#endif
    pHalFunc->RxReturn(hal_receive_info, 1);
  }
  else
  {
    pHalFunc->RxReturn(hal_receive_info, 0);
  }

  #ifdef TIATM_INST_SUPP
  psp_trace_par(ATM_DRV_RX_EXIT, bRet);
  #endif

  return bRet;
}


int tn7efm_receive (void *os_dev, int ch, unsigned int packet_size,
                    void *os_receive_info, void *data)
{
  struct net_device *dev= (struct net_device *) os_dev;
  Tn7AtmPrivate *priv;
  struct sk_buff *skb;
  static unsigned int ledticks;
  static unsigned int wait_count=0;

  dgprintf (4, "tn7efm_receive\n");

#ifdef TIATM_INST_SUPP
  psp_trace_par (ATM_DRV_PKT_RX_ENTER, packet_size);
#endif

#ifdef EFM_DEBUG
  if (os_dev != mydev_efm)
    printk("tn7efm_receive: dev=%08x, expect=%08x\n", (unsigned int) dev,
      (unsigned int) mydev_efm);
#endif

  priv = netdev_priv(dev);

#ifdef EFM_DEBUG
  if (priv != mypriv)
    printk("tn7efm_receive: priv=%08x, expect=%08x\n", (unsigned int) priv, (unsigned int) mypriv);
#endif

  if (priv->lut[ch].ready == 0)
  {
    if ((wait_count++ % 1000)==0)
      printk ("%s: channel not ready %d\n", __FUNCTION__, ch);
    priv->stats.rx_errors++;
    return 1;
  }

  /*
   * assume no fragment packet for now
   */
  skb = (struct sk_buff *) os_receive_info;

  if (skb == NULL)
  {
    dgprintf (1, "received empty skb.\n");
    priv->stats.rx_errors++;
#ifdef TIATM_INST_SUPP
    psp_trace_2_par (ATM_DRV_PKT_RX_EXIT, -1, -2);
#endif
    return 1;
  }

  skb->len = packet_size;

  /*
   * skb_trim(skb,skb->len);
   */
  /*
   * skb size is incorrect for large packets > 1428 bytes ??
   */
  __skb_trim (skb, skb->len);   /* change to correct > 1500 ping when
                                 * firewall is on */

  dgprintf (3, "pushing the skb...\n");

#if 0
  //AV_26
  skb->stamp.tv_sec = xtime.tv_sec;
  skb->stamp.tv_usec = xtime.tv_nsec;   //Sorry
#endif

#ifdef EFM_DEBUG
  if (g_efm_proc_ctl & 2)
  {
    printk("\nRx:skb->len=%d, data_len=%d, truesize=%d, sizeof(skbuf)=%d, skb->data=%08x\n",
      skb->len, skb->data_len,
      skb->truesize, sizeof(struct sk_buff),
      (unsigned int) skb->data);

    if (skb->data)
      xxdump(skb->data, skb->len, "efmRx_data:");
  }
#endif


#ifdef EFM_DEBUG
  if (g_efm_proc_ctl & 0x100)
  { // validate Rx CRC32, for debug only
    unsigned int pkt_rx_crc32=0;
    unsigned char *ptail;
    ETH_CRC32 crc32;
    int drop_packet=0;

    if (skb->len > 1600)
    {
      drop_packet=1;  // don't even try to calc, it's corrupted.
      efm_stats.rx_crc32_dropped_len++; // dropped due to corrupted length
    }
    else
    {
      crc32.x = efm_calc_crc32(skb->data, skb->len-EGRESS_TRAILOR_LEN); // don't count FCS for FCS field.
      ptail = skb->tail-4; // backward 4 bytes for FCS
      pkt_rx_crc32 = (unsigned int) ptail[0] + ((unsigned int) ptail[1]<<8) +
        ((unsigned int) ptail[2]<<16) +  ((unsigned int) ptail[3]<<24);

      if (crc32.x != pkt_rx_crc32)
      {
        efm_stats.rx_crc32_dropped++;
        drop_packet = 1;
      }
    }

    if (drop_packet)
    {
        if (g_efm_proc_ctl & 0x200)
          printk("Rx dropped: crc32.x=%08x, pkt_rx_crc32=%08x\n", crc32.x, pkt_rx_crc32);
        // drop due to Rx CRC32 not matching
        priv->stats.rx_dropped++;
        dev_kfree_skb (skb);
        return 0;
    }
    else
    {
        if (g_efm_proc_ctl & 0x200)
          printk("Rx dropped: crc32.x=%08x, pkt_rx_crc32=%08x\n", crc32.x, pkt_rx_crc32);
        efm_stats.rx_crc32_undropped++;
    }
  }
#endif

   /* Write metadata, and then pass to the receive level */
  skb_trim(skb, skb->len- EGRESS_TRAILOR_LEN); // remove last 4 bytes CRC32

  skb->dev = dev;
  
//  skb->protocol = eth_type_trans(skb, dev);

  priv->stats.rx_packets++;
  priv->stats.rx_bytes += skb->len; // pkt->datalen;
  
  if (skb->dev->br_port == NULL)
  {
      skb->protocol = eth_type_trans(skb, dev);
  }
  
  netif_rx(skb);

#if defined (CONFIG_LED_MODULE) || defined (CONFIG_MIPS_AVALANCHE_COLORED_LED)
  if ((ledticks != jiffies) && (hnd_LED_0))
  {
    ledticks = jiffies;
    TN7DSL_LED_ACTION(hnd_LED_0, MOD_ADSL, DEF_ADSL_ACTIVITY);
  }
#endif

#ifdef TIATM_INST_SUPP
  psp_trace_2_par (ATM_DRV_PKT_RX_EXIT, 0, skb->truesize);
#endif

  return 0;
}


static struct net_device_stats *tn7efm_stats(struct net_device *dev)
{
  Tn7AtmPrivate *priv = netdev_priv(dev);
  return &priv->stats;
}

int tn7efm_sar_activate_chan (struct net_device *dev, int chan)
{
  Tn7AtmPrivate *priv=netdev_priv(dev);
  CHANNEL_INFO chInfo;
  HAL_FUNCTIONS *pHalFunc;
  HAL_DEVICE *pHalDev;
  int rc;

  pHalFunc = (HAL_FUNCTIONS *) priv->pSarHalFunc;
  pHalDev = (HAL_DEVICE *) priv->pSarHalDev;

  memset (&chInfo, 0xff, sizeof (chInfo));

  /*
   * channel specific
   */
  chInfo.Channel = chan;
  chInfo.Direction = 0;
        
  if (chan == 0)     
    chInfo.Priority=0;
  else if (chan == 1)     
    chInfo.Priority=1;

  chInfo.DaMask = 1; 
  /*
   * default
   */
  chInfo.PktType = PACKET_TYPE_NULL;

  chInfo.RxNumBuffers = priv->efm_sarRxBuf;
  chInfo.RxServiceMax = priv->efm_sarRxMax;
  chInfo.RxBufferOffset = 0;

  chInfo.TxNumBuffers = priv->efm_sarTxBuf;
  chInfo.TxServiceMax = priv->efm_sarTxMax;
  chInfo.TxNumQueues = TX_EFM_QUEUE_NUM;

  chInfo.RxBufSize = 1582;

  rc = pHalFunc->ChannelSetup(pHalDev, &chInfo, 0);
  if (rc)
  {
    printk ("failed to setup channel =%d with return code %d \n", chan, rc);
    return -1;
  }

  return 0;
}

static int tn7efm_ioctl(struct net_device *dev, struct ifreq *rq, int cmd)
{
  return 0;
}

/*
 * The "change_mtu" method is usually not needed.
 * If you need it, it must be like this.
 */
static int tn7efm_change_mtu(struct net_device *dev, int new_mtu)
{
  Tn7AtmPrivate *priv = (Tn7AtmPrivate *) netdev_priv(dev);
  OS_FUNCTIONS *os_funcs;

  /* check ranges */
  if ((new_mtu < 68) || (new_mtu > 1500))
      return -EINVAL;

  os_funcs= (OS_FUNCTIONS *) priv->pSarOsFunc;
  if (!os_funcs)
      return -1;

  /*
   * Do anything you need, and the accept the value
   */
//  DDA_sar_critical_on();
  os_funcs->CriticalOn();
  dev->mtu = new_mtu;
//  DDA_sar_critical_off();
  os_funcs->CriticalOff();
  return 0; /* success */
}

#if 0
static void tn7efm_tx_timeout (struct net_device *dev)
{
  Tn7AtmPrivate *priv = netdev_priv(dev);

  printk("Transmit timeout at %ld, latency %ld\n", jiffies,
                  jiffies - dev->trans_start);
#if 0 // not for now...
  /* Simulate a transmission interrupt to get things moving */
  priv->status = TN7EFM_TX_INTR;
  tn7efm_interrupt(0, dev, NULL);
#endif

  priv->stats.tx_errors++;
  netif_wake_queue(dev);
  return;
}

#endif

static void tn7efm_sar_exit(Tn7AtmPrivate *priv)
{
  HAL_FUNCTIONS *pHalFunc;
  HAL_DEVICE    *pHalDev;

  dgprintf(4, "tn7efm_sar_exit\n");

  pHalFunc = (HAL_FUNCTIONS *)priv->pSarHalFunc;
  pHalDev  = (HAL_DEVICE *)priv->pSarHalDev;

  if (pHalFunc && pHalDev)
  {
    pHalFunc->Close(pHalDev, 2);
    pHalFunc->Shutdown(pHalDev);
  }

  dgprintf(4, "tn7efm_sar_exit done\n");
}

int tn7efm_close (Tn7AtmPrivate *priv)
{
    dgprintf(4, "tn7efm_close\n");
    priv->lConnected = 0;
    tn7efm_sar_exit(priv);
    dgprintf(4, "tn7efm_close done\n");
    return 0;
}

void tn7efm_exit (void)
{
  struct net_device *dev= mydev_efm;
  Tn7AtmPrivate *priv= mypriv;

  dgprintf (4, "tn7efm_exit()\n");

  tn7efm_dev_close (dev);
  tn7efm_close (priv);

  /*
   * freeup irq's
   */
  if (priv->sar_irq)
  {
    free_irq (priv->sar_irq, priv->efm_dev);
    priv->sar_irq = 0;
  }

  unregister_netdev(dev);
  kfree(dev);

}


int tn7efm_lut_clear (struct net_device *dev, int chan)
{
  Tn7AtmPrivate *priv=netdev_priv(dev);

  memset (&priv->lut[chan], 0, sizeof (priv->lut[chan]));
  return 0;
}

int tn7efm_set_lut (struct net_device *dev, int chan)
{
  Tn7AtmPrivate *priv=netdev_priv(dev);
  int retval= -1; // err

  if (!priv->lut[chan].inuse)
  {
    priv->lut[chan].vpi = 0;
    priv->lut[chan].vci = 0;
    priv->lut[chan].chanid = chan;
    priv->lut[chan].inuse = 1;  /* claim the channel */
    priv->lut[chan].vcc = 0;
    priv->lut[chan].bClosing = 0;
    priv->lut[chan].ready = 0;
    priv->lut[chan].tx_total_bufs = TX_BUFFER_NUM;
    priv->lut[chan].tx_used_bufs[0] = 0;
    priv->lut[chan].tx_used_bufs[1] = 0;
    retval = 0; // ok
  }

  return retval;
}

#ifdef EFM_DEBUG
//-------------------------------- proc ------------------------------------------------
int tn7efm_proc_channels (char *buf, char **start, off_t offset,
                                int count, int *eof, void *data)
{
extern int disp_RCB_Queue_detail(char *buf, int count, int chan, HAL_DEVICE *pSarHalDev);
extern int disp_chan_detail(char *buf, int count, int chan,  HAL_DEVICE *pSarHalDev);
extern int disp_TCB_Queue(char *buf, int count, int chan, HAL_DEVICE *pSarHalDev, int Queue);
extern int disp_TCB_detail(char *buf, int count, void *Tcb_addr);
extern int disp_TCB_Queue_detail(char *buf, int count, int chan, HAL_DEVICE *pSarHalDev, int Queue);
  int len = 0;
  int limit = count - 80;
  int i;
  int chan=0;
  Tn7AtmPrivate *priv= mypriv; // Note: can't use *data, it point to atm_dev

  if (len <= limit)
    len += sprintf (buf + len, "Chan  Inuse   ChanID\n");
  if (len <= limit)
    len +=
      sprintf (buf + len,
               "----------------------------------------------------\n");

  for (i = 0; i < MAX_DMA_CHAN; i++)
  {
    if (len <= limit)
    {
      len += sprintf (buf + len,
                      " %02d    %05d   %05d\n",
                      i, priv->lut[i].inuse, priv->lut[i].chanid);
    }
  }

  if (len <= limit)
    len +=
      sprintf (buf + len,
               "------------------------------------------------------\n");
for (chan=0; chan<NUM_EFM_CHAN; chan++)
{
  if (len<= limit)
    len += sprintf( buf + len, "RCB:\n");
  if (len<= limit)
    len+= disp_RCB_Queue_detail(buf + len, count, chan, priv->pSarHalDev);
  if (len<= limit)
    len += disp_chan_detail(buf + len, count, chan, priv->pSarHalDev);

  if (len<= limit)
    len += sprintf( buf + len, "TCB:\n");

  if (len<= limit)
    len += sprintf(buf + len, "chan  TCBstart TxActQueueCount TcbPool TxActQueueHead TxActQueueTail TxActive\n");

  if (len<= limit)
    len += disp_TCB_Queue( buf + len, count, chan, priv->pSarHalDev, 0);

  if (len<= limit)
    len += disp_TCB_Queue( buf + len, count, chan, priv->pSarHalDev, 1);

  if (len<= limit)
  {
    len+= sprintf(buf + len, "TCB Queue0:\n");
    len+= disp_TCB_Queue_detail(buf + len, count, chan, priv->pSarHalDev, 0);
  }
}
  if (len<= limit)
  {
    len+= sprintf(buf + len, "availTx=%d\n", priv->available_tx_bufs[0]);
  }

  return len;
}
#endif



#define EFM_CELL_SIZE 65

// RX EFM_TC State Structure, this structure is maintained in datapump
typedef struct
{
  int    rx_state;
  UINT32 sync_bytes_missed;        // Total sync bytes missed, where expected
  UINT32 stat_oos_count;
  UINT32 idle_cell_generated;       // Number of generated ATM idle cell
  UINT32 rx_good_cell_count;       // Number of received good EFM cell
  UINT32 temp_debugflag;
  UINT8 rx_sync_histogram[EFM_CELL_SIZE];
  UINT8 bytes_needed;             // Number of bytes needed to complete present cell
  UINT8 rx_histogram_index;       // Index into rx sync byte histogram
  UINT8 tc_link_state;
} rx_efm_tc_state_type;

// TX EFM_TC State Structure, this structure is maintained in datapump
typedef struct
{
  UINT8  ATM_aligned;                     // Boolean variable indicating the state of ATM cell alignment
  UINT8  PTM_OOSCell;                     // Boolean variable indicating the generation of EFM all idle out-of-sync cell
  SINT16 bytes_remaining_in_cell;         // 8 bit should be enough
  UINT32 stat_idle_gen_count;             // Number of all idle cells generated by EFM-TC
  UINT32 stat_oos_idle_gen_count;         // Number of all idle OOS cells generated by EFM-TC
  UINT32 tx_efm_cell_count;               // Number of transmitted EFM cell, including OOS cells
} tx_efm_tc_state_type;


#define SAVE_SIZE 4000
struct sk_buff *skbuf_save[SAVE_SIZE];

int tn7efm_proc_info (char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
  Tn7AtmPrivate *priv= mypriv; // Note: can't use *data, it point to atm_dev
  PITIDSLHW_T pIhw = priv->pIhw;
  int len = 0;
  int limit = count - 80;
  unsigned int rx_efm_stateAddr_offset[2]={7, 21};
  unsigned int tx_efm_stateAddr_offset[2]={7, 22};
  tx_efm_tc_state_type *rx_efmStat_p;
  tx_efm_tc_state_type *tx_efmStat_p;
  tx_efm_tc_state_type  tx_efmStat_s;
  rx_efm_tc_state_type  rx_efmStat_s;
  struct {
    unsigned char *str;
  } pre_emption_tbl[]={{"not"},{"DS"},{"US"},{"DS/US"}};

#ifdef EFM_DEBUG
  volatile unsigned int *GPR= (volatile unsigned int *) (0xa3000000+ 0x20);
//  volatile unsigned int *atm_cell_count= (volatile unsigned int *) (0xa3000000+0x8050);
  unsigned int gpr3_atm1, gpr3_atm2, gpr2_atmall, gpr2_rx_good;
  unsigned int tmp;
#endif

  len += sprintf(buf + len, "curr_TC_mode=%d, EFM_mode=%d\nefm_send_count=%d\n",
      priv->curr_TC_mode, priv->EFM_mode, efm_stats.send_count);

  len += sprintf(buf + len, "curr_PTM_pre_emption=%s enabled\n",pre_emption_tbl[pIhw->PTM_pre_emption].str);
       
  if (priv->curr_TC_mode != TC_MODE_PTM)
    return len;   // Not in PTM mode

  len+= sprintf(buf + len, "efmDevInt(Tot/TH/TL/R/STS)=%d/%d/%d/%d/%d\n",
    efm_stats.DevInt_count, efm_stats.DevInt_TH_count,
    efm_stats.DevInt_TL_count, efm_stats.DevInt_Rx_count, efm_stats.DevInt_STS_count);

//  len += sprintf(buf + len, "Tx: pkts/bytes/dropped/errors=%d/%d/%d/%d\n",
//    priv->stats.tx_packets, priv->stats.tx_bytes, priv->stats.tx_dropped, priv->stats.tx_errors);

//  len += sprintf(buf + len, "Rx: pkts/bytes/dropped/errors=%d/%d/%d/%d\n",
//    priv->stats.rx_packets, priv->stats.rx_bytes, priv->stats.rx_dropped, priv->stats.rx_errors);

  len += sprintf(buf + len, "efm_queue_packet_to_sar_count=%d\n", efm_stats.queue_packet_to_sar_count);

#ifdef  EFM_DEBUG
  len += sprintf(buf + len, "g_rx_crc32_dropped/len/undropped/tot=%d/%d/%d/%d\n",
    efm_stats.rx_crc32_dropped,
    efm_stats.rx_crc32_dropped_len,
    efm_stats.rx_crc32_undropped,
    efm_stats.rx_crc32_dropped+efm_stats.rx_crc32_undropped+efm_stats.rx_crc32_dropped_len);
#endif

  rx_efmStat_p = (tx_efm_tc_state_type *) dslhal_support_readInternalOffset(pIhw, 2, rx_efm_stateAddr_offset);
  tx_efmStat_p = (tx_efm_tc_state_type *) dslhal_support_readInternalOffset(pIhw, 2, tx_efm_stateAddr_offset);

#ifdef EFM_DEBUG
  len+= sprintf(buf+ len, "rx_efmStat_p=%08x, tx_efmStat_p=%08x\n", (unsigned int) rx_efmStat_p,
    (unsigned int) tx_efmStat_p);
#endif

  if (pIhw->AppData.bState == RSTATE_SHOWTIME)
  {
    dslhal_support_blockRead((void *) rx_efmStat_p, &rx_efmStat_s, sizeof(rx_efmStat_s));
    dslhal_support_blockRead((void *) tx_efmStat_p, &tx_efmStat_s, sizeof(tx_efmStat_s));
  }
  else{
    memset(&rx_efmStat_s, 0, sizeof(rx_efmStat_s));
    memset(&tx_efmStat_s, 0, sizeof(tx_efmStat_s));
  }

  len+= sprintf (buf+len, "Rx:\n");
  len+= sprintf (buf+len, "  rx_state: %d\n", rx_efmStat_s.rx_state);
  len+= sprintf (buf+len, "  sync_bytes_missed= %d\n", rx_efmStat_s.sync_bytes_missed);
  len+= sprintf (buf+len, "  stat_oos_count=%d\n", rx_efmStat_s.stat_oos_count);
  len+= sprintf (buf+len, "  idle_cell_generated= %d\n", rx_efmStat_s.idle_cell_generated);
  len+= sprintf (buf+len, "  rx_good_cell_count= %d\n", rx_efmStat_s.rx_good_cell_count);
  len+= sprintf (buf+len, "  temp_debugflag= %d\n", rx_efmStat_s.temp_debugflag);
  len+= sprintf (buf+len, "  tc_link_state= %d\n", rx_efmStat_s.tc_link_state);

#ifdef EFM_DEBUG
  tmp = GPR[3];

gpr3_atm1=tmp & 0xFFFF;
gpr3_atm2=(tmp>>16) & 0xFFFF;

tmp = GPR[2];
gpr2_atmall= tmp & 0xFFFF;
gpr2_rx_good= (tmp>>16) & 0xFFFF;
        len+= sprintf(buf+len," pdsp:GPR0-7 = %d %d %d %d %d %d %d %d\n",
        GPR[0], GPR[1], GPR[2], GPR[3], GPR[4], GPR[5], GPR[6], GPR[7]);
#endif

  if (len <= limit)
    len+= sprintf (buf+len, "Tx:\n");
  if (len <= limit)
    len+= sprintf (buf+len, "  ATM_aligned=%d\n", tx_efmStat_s.ATM_aligned);
  if (len <= limit)
    len+= sprintf (buf+len, "  PTM_OOSCell=%d\n", tx_efmStat_s.PTM_OOSCell);
  if (len <= limit)
    len+= sprintf (buf+len, "  bytes_remaining_in_cell=%d\n",   tx_efmStat_s.bytes_remaining_in_cell);
  if (len <= limit)
    len+= sprintf (buf+len, "  stat_idle_gen_count=%d\n", tx_efmStat_s.stat_idle_gen_count);
  if (len <= limit)
    len+= sprintf (buf+len, "  stat_oos_idle_gen_count=%d\n", tx_efmStat_s.stat_oos_idle_gen_count);
  if (len <= limit)
    len+= sprintf (buf+len, "  tx_efm_cell_count= %d\n", tx_efmStat_s.tx_efm_cell_count);

  return len;
}


int tn7efm_proc_ctrl_read (char *buf, char **start, off_t offset,
                                 int count, int *eof, void *data)
{
  Tn7AtmPrivate *priv= mypriv;
  int len = 0;
  int limit = count - 80;

  len+= sprintf(buf+len, "efm_ctl=0x%x\n", priv->efm_ctl);
  len+= sprintf(buf+len, "efm_proc_ctl=0x%02x\n", g_efm_proc_ctl);
#ifdef EFM_DEBUG
  len+= sprintf(buf+len, "DbgLevel=%d\n", __guDbgLevel);
  len+= sprintf(buf+len, "dsl_current_sync=%d (%s)\n",
    g_dsl_current_sync, g_dsl_current_sync ? "sync": "unsync");
  len+= sprintf(buf+len, "dsl_unsync_count=%u\n", g_dsl_unsync_count);
#endif
  if (len <= limit)
    len+= sprintf(buf+len, "tcmode=%d\n", priv->curr_TC_mode);

  return len;
}

int tn7efm_proc_ctrl_write (struct file *fp, const char *buf, unsigned long count, void *data)
{
  Tn7AtmPrivate *priv= mypriv;
  char local_buf[32];
  char token1[20], token2[20];
  unsigned int val;

  if(count > 30)
  {
    printk("Error : Buffer Overflow\n");
    return -1;
  }

  copy_from_user(local_buf,buf,count);
  local_buf[count-1] = '\0';

  if (2 != sscanf(local_buf, "%s %s", token1, token2))
  {
    printk("wrong format\nusage: echo token1 token2 > avsar_efm_ctl\n");
    return -1;
  }

  val = os_atoih(token2);

  if (!strcmp(token1,"efm_proc_ctl"))
  {
    printk("efm_proc_ctl<-0x%02x\n", val);
    g_efm_proc_ctl = val;
  }
#ifdef DEBUG_BUILD
  else if (!strcmp(token1,"DbgLevel"))
  {
    printk("DbgLevel<-%d\n", val);
    __guDbgLevel = val;
  }
#endif
#ifdef EFM_DEBUG
  else if (!strcmp(token1,"dsl_unsync_count"))
  {
    g_dsl_current_sync=0;
    g_dsl_unsync_count=0;
    printk("dsl_unsync_count<-0\n");
    printk("dsl_current_sync<-0\n");
  }
#endif
  else if (!strcmp(token1, "tcmode"))
  {
    if ((val==TC_MODE_ATM) || (val==TC_MODE_PTM))
    {
      printk("tcmode<-%d\n", val);
      priv->target_TC_mode = val;
      tasklet_hi_schedule (&priv->tn7efm_tasklet);
    }
  }

  return count;
}

#ifdef EFM_DEBUG
#ifndef isprint
#define isprint(a) ((a >=' ')&&(a <= '~'))
#endif

void xxdump( unsigned char*  cp, int  length, char*  prefix )
{
    int col, count;
    u_char prntBuf[120];
    u_char*  pBuf = prntBuf;
    count = 0;

    while(count < length){
        pBuf += sprintf( pBuf, "%s", prefix );
        for(col = 0;count + col < length && col < 16; col++){
            if (col != 0 && (col % 4) == 0)
                pBuf += sprintf( pBuf, " " );
            pBuf += sprintf( pBuf, "%02X ", cp[count + col] );
        }
        while(col++ < 16){      /* pad end of buffer with blanks */
            if ((col % 4) == 0)
                sprintf( pBuf, " " );
            pBuf += sprintf( pBuf, "   " );
        }
        pBuf += sprintf( pBuf, "  " );
        for(col = 0;count + col < length && col < 16; col++){
            if (isprint((int)cp[count + col]))
                pBuf += sprintf( pBuf, "%c", cp[count + col] );
            else
                pBuf += sprintf( pBuf, "." );
                }
        sprintf( pBuf, "\n" );
        // SPrint(prntBuf);
        printk(prntBuf);
        count += col;
        pBuf = prntBuf;
    }

}  /* close xdump(... */

#endif

int tn7efm_sar_init(struct net_device *dev)
{
  Tn7AtmPrivate *priv = netdev_priv(dev);
  extern void efm_SetDevStateConnected(HAL_DEVICE *HalDev);
  int retCode;
  HAL_FUNCTIONS *pHalFunc;
  HAL_DEVICE    *pHalDev;
  OS_FUNCTIONS  *pOsFunc= priv->pSarOsFunc;

  dgprintf(4, "tn7efm_sar_init\n");

#if 0
  pOsFunc = (OS_FUNCTIONS  *)kmalloc(sizeof(OS_FUNCTIONS), GFP_KERNEL);

  priv->SarOsFunc = (void *)pOsFunc;
#endif

  /* init sar os call back functions */
  retCode = tn7efm_sar_init_module(pOsFunc);

  if (retCode != 0) /* error */
  {
    printk ("Failed to init EFMSAR OS Functions\n");
    return (1);
  }

  if (!priv->efm_initialized)
    efmcrc_init_crcTable();

  // Assuming ATM has initialize these
  pHalFunc = priv->pSarHalFunc;
  pHalDev = priv->pSarHalDev;

  /* Probe for the Device to get hardware info from driver */
  efm_SetDevStateConnected(pHalDev);  // pHalDev->State = enConnected;
  retCode = pHalFunc->Probe(pHalDev);

  if (retCode !=0)
  {
    printk ("EFM SAR hal probing error.\n");
    return (1);
  }

  /* init sar hal */
  retCode = pHalFunc->Init(pHalDev);
  if (retCode != 0) /* error */
  {
    printk ("pHalFunc->Init failed. err code =%d\n", retCode);
    return (1);
  }

  /* open hal module */
  retCode = pHalFunc->Open(pHalDev);
  if (retCode != 0) /* error */
  {
     printk ("pHalFunc->open failed, err code: %d\n",retCode );
     return (1);
  }

  dgprintf(4, "tn7efm_sar_init done\n");
  return 0;
}


// Only overwrite the different functions
int tn7efm_sar_init_module(OS_FUNCTIONS *os_funcs)
{
  dgprintf(4, "tn7efm_init_module\n");
  if( os_funcs == 0 )
  {
    return(-1);
  }
#if 0 // reuse values from ATM
  os_funcs->Control                  = tn7sar_control;
  os_funcs->CriticalOn               = tn7sar_critical_on;
  os_funcs->CriticalOff              = tn7sar_critical_off;
  os_funcs->DataCacheHitInvalidate   = tn7atm_data_invalidate;
  os_funcs->DataCacheHitWriteback    = tn7atm_data_writeback;
  os_funcs->DeviceFindInfo           = tn7sar_find_device;
  os_funcs->DeviceFindParmUint       = tn7sar_get_device_parm_uint;
  os_funcs->DeviceFindParmValue      = tn7sar_get_device_parm_value;
  os_funcs->Free                     = tn7sar_free;
  os_funcs->FreeRxBuffer             = tn7sar_free_buffer;
  os_funcs->FreeDev                  = tn7sar_free_dev;
  os_funcs->FreeDmaXfer              = tn7sar_free_dma_xfer;
  os_funcs->IsrRegister              = tn7sar_sarhal_isr_register;
  os_funcs->IsrUnRegister            = tn7sar_isr_unregister;
  os_funcs->Malloc                   = tn7sar_malloc;
  os_funcs->MallocRxBuffer           = tn7sar_malloc_rxbuffer;
  os_funcs->MallocDev                = tn7sar_malloc_dev;
  os_funcs->MallocDmaXfer            = tn7sar_malloc_dma_xfer;
  os_funcs->Memset                   = tn7sar_memset;
  os_funcs->Printf                   = printk;
#endif

  os_funcs->Receive                  = tn7efm_sar_receive;
  os_funcs->SendComplete             = tn7efm_send_complete;

#if 0 // reuse values from ATM
  os_funcs->Strcmpi                  = strcmp;
  os_funcs->Sprintf                  = sprintf;
  os_funcs->Strlen                   = strlen;
  os_funcs->Strstr                   = strstr;
  os_funcs->Strtoul                  = tn7sar_strtoul;
  os_funcs->TeardownComplete         = tn7sar_teardown_complete;
#endif

  dgprintf(4, "tn7efm_init_module done\n");
  return(0);
}


/*
 * tn7efm_showtime_init:
 *    This function is invoked once right after each showtime.
*/
void tn7efm_showtime_init (void)
{
  memset(&efm_stats, 0, sizeof(efm_stats)); // clean up the efm_stats
}